"""
Test that the library converts a set of test scripts correctly.
"""
import os
import pytest

from .. import modernize


TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


def load_test_scripts(path):
    """
    Utility function to load all scripts from the given folder and return them
    in a dict.
    """
    scripts = dict()
    for fname in os.listdir(path):
        if os.path.splitext(fname)[-1] == '.sh' and fname[0] != '~':
            with open(os.path.join(path, fname)) as script:
                scripts[fname] = script.read()
    return scripts


@pytest.fixture(scope='module')
def classic():
    """
    Scripts in classic mode.
    """
    return load_test_scripts(os.path.join(TEST_DATA_DIR, 'classic'))


@pytest.fixture(scope='module')
def modern():
    """
    Scripts in modern mode corresponding to the same scripts in classic mode.
    """
    return load_test_scripts(os.path.join(TEST_DATA_DIR, 'modern'))


def test_integration(classic, modern):
    'Compare converted scripts to a reference of what they should be.'
    assert set(classic.keys()) == set(modern.keys()), \
        'Different number of scripts in classic and modern data dirs'
    for script in classic:
        modernized = modernize(classic[script])
        assert modernized == modern[script], \
            'Failed {}.\n\nModernized:\n\n{}\n\nExpected:\n\n{}\n\n'.format(
                script, modernized, modern[script])



"""
Base package for the gmtmodernize program.
"""
from ._version import get_versions
from .conversion import modernize


__version__ = get_versions()['version']
del get_versions


def test(doctest=True, verbose=True, coverage=False):
    """
    Run the test suite.

    Uses `py.test <http://pytest.org/>`__ to discover and run the tests. If you
    haven't already, you can install it with `conda
    <http://conda.pydata.org/>`__ or `pip <https://pip.pypa.io/en/stable/>`__.

    Parameters:

    * doctest : bool
        If ``True``, will run the doctests as well (code examples that start
        with a ``>>>`` in the docs).
    * verbose : bool
        If ``True``, will print extra information during the test run.
    * coverage : bool
        If ``True``, will run test coverage analysis on the code as well.
        Requires ``pytest-cov``.

    Raises:

    * ``AssertionError`` if pytest returns a non-zero error code indicating
      that some tests have failed.

    """
    import pytest
    args = []
    if verbose:
        args.append('-vv')
    if coverage:
        args.append('--cov=gmtmodernize')
        args.append('--cov-report=term-missing')
    if doctest:
        args.append('--doctest-modules')
    args.append('--pyargs')
    args.append('gmtmodernize')
    status = pytest.main(args)
    assert status == 0, "Some tests have failed."



"""Convert GMT shell scripts from classic to modern mode.

Prints the converted modern mode script to standard output (STDOUT) when
converting a single script.

Usage:
    gmtmodernize SCRIPT
    gmtmodernize --recursive [--quiet] FOLDER_CLASSIC FOLDER_MODERN
    gmtmodernize --help
    gmtmodernize --version

Arguments:
    SCRIPT          Classic mode script to convert.
    FOLDER_CLASSIC  Folder with classic mode scripts (can have multiple
                    sub-folders).
    FOLDER_MODERN   Name of output folder with converted modern mode scripts.
                    Mirrors the folder structure of FOLDER_CLASSIC and copies
                    all non-script files.

Options:
    -r --recursive  Recursively transverse a folder structure with GMT scripts
                    and other files instead of converting a single file.
                    Creates a new folder with the same structure and non-script
                    files copied over, plus the converted GMT scripts.
    --quiet         Don't print any text to STDERR while processing.
    -h --help       Show this help message and exit.
    --version       Show the version and exit.

Examples:

    Convert a single GMT script to modern mode:

        $ gmtmodernize classic_script.sh > modern_script.sh

    Convert a folder with GMT scripts, data files, etc, (optionally inside
    multiple sub-folders):

        $ gmtmodernize -r gmt_classic_scripts/ gmt_modern_scripts/

    This will create a folder 'gmt_modern_scripts' with the same sub-folders
    and non-script files in 'gmt_classic_scripts' but with the scripts
    converted to modern mode.
"""
import os
import sys
import shutil
from functools import partial
from docopt import docopt

from . import __version__
from . import modernize


def is_gmt_script(fname):
    """
    Check if a given file is a GMT script.
    """
    return os.path.splitext(fname)[-1] == '.sh'


def find_gmt_scripts(directory):
    """
    Find all GMT scripts in the given directory (including subdirectories).
    """
    gmt_scripts = []
    for base, _, fnames in os.walk(directory):
        full_fnames = (os.path.join(base, f) for f in fnames)
        gmt_scripts.extend(f for f in full_fnames if is_gmt_script(f))
    return gmt_scripts


def mirror_directory(original, target, ignore):
    """
    Mirror the original directory structure.

    Creates a new 'target' folder and copy all files and subfolders from
    'original', except the files in the 'ignore' list.
    """
    ignore = set(ignore)
    for base, _, fnames in os.walk(original):
        base_target = base.replace(original, target)
        os.mkdir(base_target)
        to_copy = set(os.path.join(base, f) for f in fnames).difference(ignore)
        for fname in to_copy:
            shutil.copy(fname, base_target)


def main():
    """
    Entry point for the command line interface.
    """
    args = docopt(__doc__, version=__version__)

    if args['--quiet'] or not args['--recursive']:

        def echo(*args, **kwargs):
            return None

    else:
        echo = partial(print, file=sys.stderr)

    if args['SCRIPT'] is not None:
        gmt_scripts = [args['SCRIPT']]
        output_names = [None]

        def save_output(content, *args, **kwargs):
            print(content)

    else:
        input_dir = os.path.normpath(args['FOLDER_CLASSIC'])
        output_dir = os.path.normpath(args['FOLDER_MODERN'])
        if os.path.exists(output_dir):
            raise RuntimeError("Target directory '{}' already exists.".format(
                output_dir))
        echo("Scanning '{}' for GMT scripts...".format(input_dir))
        gmt_scripts = find_gmt_scripts(input_dir)
        if len(gmt_scripts) < 1:
            raise RuntimeError("Didn't find any GMT scripts in '{}'.".format(
                input_dir))
        else:
            echo("Found {} GMT scripts:".format(len(gmt_scripts)))
            for fname in gmt_scripts:
                echo("  {}".format(fname))

        echo("Copying folder structure and non-script files to '{}'.".format(
            output_dir))
        mirror_directory(original=input_dir,
                         target=output_dir,
                         ignore=gmt_scripts)

        output_names = [script.replace(input_dir, output_dir)
                        for script in gmt_scripts]

        def save_output(content, fname, original_fname):
            with open(fname, 'w') as outputfile:
                outputfile.write(content)
            # Copy the file permissions to the modern script
            shutil.copymode(original_fname, fname)

    echo("Converting scripts to modern mode:")
    for classic_fname, modern_fname in zip(gmt_scripts, output_names):
        with open(classic_fname) as inputfile:
            classic_script = inputfile.read()
        modern_script = modernize(classic_script)
        save_output(modern_script, modern_fname, classic_fname)
        echo('  {}'.format(modern_fname))

    echo("Done")



"""
Functions to convert a classic script into a modern one.

The main functionality is exposed through the ``modernize`` function.
"""
import re
import os


# List of GMT modules that generate Postscript output
PS_MODULES = """
gmtlogo
grdcontour
grdimage
grdvector
grdview
psbasemap
psclip
pscoast
pscontour
pshistogram
psimage
pslegend
psmask
psrose
psscale
pssolar
pstext
pswiggle
psxyz
psxy
pscoupe
psmeca
pspolar
pssac
psvelo
mgd77track
pssegyz
pssegy
""".split()


def modernize(script):
    """
    Convert a script from classic to modern mode.

    Parameters
    ----------
    script : str
        Classic mode script.

    Returns
    -------
    str
        Script converted to modern mode.

    """
    modern = []
    script = script.split('\n')[:-1]

    # Copy the header comments and add a command to set the mode
    # to modern after the header
    for line in script:
        line = line.rstrip()
        if not line or line[0] != '#':
            break
        modern.append(line)
    last_line = len(modern)

    # Look for the "ps" variable definition and extract the name
    ps_var_line = None
    for l, line in enumerate(script[last_line:]):
        line = line.strip()
        # Check if this line defines a .ps file name variable
        ps_var_def = re.findall(r'^ps=.+\.ps$', line)
        if ps_var_def:
            assert len(ps_var_def) == 1, \
                "Found more than 1 ps variable in line '{}'".format(line)
            ps_definition = ps_var_def[0]
            line = os.path.splitext(ps_definition)[0]
            modern.append('')
            modern.append(line)
            ps_var_line = last_line + l
            break
    if ps_var_line is not None:
        script.pop(ps_var_line)
        begin_args = " $ps ps"
    else:
        begin_args = ""

    modern.append('')
    modern.append('gmt begin{}'.format(begin_args))
    modern.append('')

    # Parse the rest of the script
    for line in script[last_line:]:
        line = line.rstrip()

        # Check if redirecting to the $ps variable and strip it out
        redirect_to_ps = re.findall(r'.+>+ *\$ps(?: +|$)', line)
        if redirect_to_ps:
            assert len(redirect_to_ps) == 1, \
                "Found more than 1 ps redirection in line '{}'".format(
                    line)
            redirection = re.findall(r'>+ *\$ps(?: +|$)', line)[0]
            line = line.replace(redirection, '').strip()

        # Remove -O -K -R -J
        okrj = re.findall(r'(?:(?<= )|^)-[KORJ](?: +|$)', line)
        if okrj:
            for match in okrj:
                line = line.replace(match, '')
            line = line.strip()

        modern.append(line)

    modern.append('')
    modern.append('gmt end')
    modern.append('')

    # Remove duplicate blank lines
    modern = [line for l, line in enumerate(modern)
              if l == 0 or not (line == '' and modern[l - 1] == '')]

    return '\n'.join(modern)



"""
Test components of the command line program.
"""
import os
import shutil

from ..cli import find_gmt_scripts, mirror_directory, is_gmt_script


TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


def test_find_gmt_scripts():
    """
    Check if find_gmt_scripts finds all files in the 'classic' folder of the
    test data.
    """
    folder = os.path.join(TEST_DATA_DIR, 'classic')
    existing_scripts = set(os.path.join(folder, f) for f in os.listdir(folder))
    scripts_found = set(find_gmt_scripts(folder))
    assert scripts_found == existing_scripts


def test_is_gmt_script():
    """
    Check if identifying scripts correctly.
    """
    scripts = 'bla.sh dir/meh.sh'.split()
    not_scripts = 'meh bla bla/foo.txt foo.abc'.split()
    for script in scripts:
        assert is_gmt_script(script), 'Failed for {}'.format(script)
    for not_script in not_scripts:
        assert not is_gmt_script(not_script), \
            'Failed for {}'.format(not_script)


def test_mirror_directory():
    """
    Check if the correct directory structure is copies (excluding script files)
    """
    pjoin = os.path.join
    folder = pjoin(TEST_DATA_DIR, 'mirror_directory')
    target = folder.replace('mirror_directory', 'test_mirrored')
    # Copy the entire directory
    try:
        mirror_directory(folder, target, [])
        true_files = [
            "not_script.xyz",
            "script.sh",
            pjoin("level1", "not_script.txt"),
            pjoin("level1", "level2", "not_script.txt"),
            pjoin("level1", "level2", "script.sh"),
            pjoin("subfolder", "not_script.foo"),
            pjoin("subfolder", "script.sh"),
        ]
        should_contain = [pjoin(target, f) for f in true_files]
        contains = [pjoin(base, f)
                    for base, _, files in os.walk(target)
                    for f in files]
        assert set(contains) == set(should_contain), "Failed for all files"
    finally:
        shutil.rmtree(target)
    # Ignore a few files
    ignore = [pjoin(folder, f)
              for f in ['script.sh',
                        pjoin('level1', 'level2', 'script.sh'),
                        pjoin('subfolder', 'script.sh')]]
    try:
        mirror_directory(folder, target, ignore)
        true_files = [
            "not_script.xyz",
            pjoin("level1", "not_script.txt"),
            pjoin("level1", "level2", "not_script.txt"),
            pjoin("subfolder", "not_script.foo"),
        ]
        should_contain = [pjoin(target, f) for f in true_files]
        contains = [pjoin(base, f)
                    for base, _, files in os.walk(target)
                    for f in files]
        assert set(contains) == set(should_contain), "Failed for ignored files"
    finally:
        shutil.rmtree(target)



from setuptools import setup, find_packages
import versioneer

# PACKAGE METADATA
# #############################################################################
NAME = 'gmtmodernize'
FULLNAME = 'gmtmodernize'
DESCRIPTION = "Tool to convert GMT scripts to the new modern execution mode"
AUTHOR = "Leonardo Uieda"
AUTHOR_EMAIL = 'leouieda@gmail.com'
MAINTAINER = AUTHOR
MAINTAINER_EMAIL = AUTHOR_EMAIL
VERSION = versioneer.get_version()
CMDCLASS = versioneer.get_cmdclass()
with open("README.rst") as f:
    LONG_DESCRIPTION = ''.join(f.readlines())
PACKAGES = find_packages(exclude=['doc', 'ci', 'test', 'example'])
LICENSE = "BSD License"
URL = "https://github.com/GenericMappingTools/gmtmodernize"
PLATFORMS = "Any"
SCRIPTS = []
PACKAGE_DATA = {
    'gmtmodernize.tests': [
        'data/classic/*',
        'data/modern/*',
        'data/mirror_directory/*',
        'data/mirror_directory/subfolder/*',
        'data/mirror_directory/level1/*',
        'data/mirror_directory/level1/level2/*',
    ],
}
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "Intended Audience :: Education",
    "Topic :: Scientific/Engineering",
    "Topic :: Software Development :: Libraries",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "License :: OSI Approved :: {}".format(LICENSE),
]
KEYWORDS = ''
ENTRY_POINTS = {
    'console_scripts': ['gmtmodernize=gmtmodernize.cli:main', ]
}

# DEPENDENCIES
# #############################################################################
INSTALL_REQUIRES = [
    'docopt',
]

if __name__ == '__main__':
    setup(name=NAME,
          fullname=FULLNAME,
          description=DESCRIPTION,
          long_description=LONG_DESCRIPTION,
          version=VERSION,
          author=AUTHOR,
          author_email=AUTHOR_EMAIL,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          license=LICENSE,
          url=URL,
          platforms=PLATFORMS,
          scripts=SCRIPTS,
          packages=PACKAGES,
          package_data=PACKAGE_DATA,
          classifiers=CLASSIFIERS,
          keywords=KEYWORDS,
          install_requires=INSTALL_REQUIRES,
          cmdclass=CMDCLASS,
          entry_points=ENTRY_POINTS)
